from eth_vertigo.test_runner import Runner
from eth_vertigo.test_runner.file_editor import FileEditor
from eth_vertigo.test_runner.truffle.truffle_tester import TruffleTester
from eth_vertigo.test_runner.exceptions import EquivalentMutant
from eth_vertigo.core import Mutation, MutationResult
from typing import Generator

from pathlib import Path
from tempfile import mkdtemp
from distutils.dir_util import copy_tree
import shutil
from typing import Dict


def _make_temp_truffle_directory(original_dir: str):
    td = mkdtemp()
    copy_tree(original_dir, td, preserve_symlinks=1)
    return td


def _clean_build_directory(project_path: str):
    build_dir = Path(project_path) / "build"
    if build_dir.is_dir():
        shutil.rmtree(build_dir)


def _set_reporter(directory: str):
    config = Path(directory) / "truffle.js"
    if not config.is_file():
        config = Path(directory) / "truffle-config.js"
    content = config.read_text("utf-8")
    content += "\nmodule.exports.mocha = {reporter: \"json\"};\n"
    content += "\nmodule.exports.solc = {optimizer: { enabled: true, runs: 200}};\n"
    config.write_text(content, "utf-8")


def _rm_temp_truffle_directory(temp_dir: str):
    shutil.rmtree(temp_dir)


def _apply_mutation(mutation: Mutation, working_directory):
    target_file_name = working_directory + '/' + mutation.relative_path
    FileEditor.edit(target_file_name, mutation.location, mutation.value)


class TruffleRunner(Runner):

    def __init__(self, project_directory, truffle_tester: TruffleTester):
        self.project_directory = project_directory
        self.truffle_tester = truffle_tester

    @property
    def tests(self) -> Generator[str, None, None]:
        raise NotImplementedError

    def run_tests(self, coverage: bool = False, mutation: Mutation = None, timeout=None, network: str = None, original_bytecode: Dict[str, str] = None) -> dict:
        """
        Runs all the tests in the truffle project in a clean environment
        :param coverage: Whether to run the tests with coverage
        :param mutation: List indicating edits that need to be performed on mutator files
        :return: Test results
        """
        if coverage:
            raise NotImplementedError

        temp_dir = _make_temp_truffle_directory(self.project_directory)
        _clean_build_directory(temp_dir)
        _set_reporter(temp_dir)

        if mutation:
            _apply_mutation(mutation, temp_dir)
        try:
            if original_bytecode is not None and original_bytecode != {}:
                if self.truffle_tester.check_bytecodes(temp_dir, original_bytecode):
                    raise EquivalentMutant
            result = self.truffle_tester.run_test_command(temp_dir, timeout=timeout, network_name=network)
        finally:
            _rm_temp_truffle_directory(temp_dir)

        return result

    def run_test(self, name: str, coverage: bool = False):
        """
        Runs test with specific name (useless for now since truffle doesn't really support this
        This command will run all the tests to get results!
        :param name: Name of the test to run
        :param coverage: Whether to run the test with coverage
        :return: test result for the requested test
        """
        test_result = self.run_tests()
        if name in test_result.keys():
            return test_result[name]
        return None



from typing import Tuple


class FileEditor:
    @staticmethod
    def edit(target_file_name: str, edit_location: Tuple[int, int, int], replacement: str) -> None:
        """
        Rewrites the edit location in the target file with the replacement
        :param target_file_name: Name of the file to apply the rewrite to
        :param edit_location: Location to rewrite
        :param replacement: Replacement for location
        """
        if edit_location[0] <= 0:
            raise ValueError("Cannot edit in the negative space")

        with open(target_file_name, 'r') as file:
            content = file.read()

        with open(target_file_name, 'w') as file:
            new_content = content[:edit_location[0]] \
                          + replacement + \
                          content[edit_location[0] + edit_location[1]:]
            file.write(new_content)



class TestResult:
    def __init__(self, title: str, full_title: str, duration: int, success: bool):
        """ Initialize test result object

        :param title: Title of the test
        :param full_title: Full title of the test
        :param duration: Duration of the test
        :param success: Whether the test has succeeded
        """
        self.title = title
        self.full_title = full_title
        self.duration = duration
        self.success = success



from eth_vertigo.test_runner.truffle.truffle_tester import TruffleTester
from eth_vertigo.core.truffle.truffle_compiler import TruffleCompiler

from eth_vertigo.test_runner.exceptions import TestRunException, TimedOut
from eth_vertigo.test_runner import TestResult
from json import loads, JSONDecodeError
from subprocess import Popen, TimeoutExpired
from tempfile import TemporaryFile
from pathlib import Path
from typing import Dict, Union
from loguru import logger
import json
import re

swarm_hash_regex = re.compile("((a165)(.*)(5820)[a-f0-9]{64}(0029))$")


class Truffle(TruffleTester, TruffleCompiler):
    """ Truffle interface object, deals with the ugly commandline details"""

    def __init__(self, truffle_location: str = "truffle") -> None:
        """ Initializes a new truffle object

        :param truffle_location: Location where the truffle cli can be found
        """
        self.truffle_location = truffle_location

    def run_compile_command(self, working_directory: str) -> None:
        command = [
            self.truffle_location, 'compile'
        ]

        with TemporaryFile() as stdin, TemporaryFile() as stdout:
            stdin.seek(0)
            proc = Popen(command, stdin=stdin, stdout=stdout, cwd=working_directory)
            proc.wait()
            stdout.seek(0)
            output = stdout.read()

        split = output.decode('utf-8').split("\n")

        errors = []
        for line in split:
            if line.startswith("Error"):
                errors.append(line)

        if errors:
            raise Exception("Encountered compilation error: \n" + "\n".join(errors))

    def run_test_command(self, working_directory: str, file_name: str = None, network_name: str = None, timeout=None) -> Union[Dict[str, TestResult], None]:
        command = [
            self.truffle_location, 'test'
        ]

        if file_name:
            command.append(file_name)
        if network_name:
            command += ['--network', network_name]

        with TemporaryFile() as stdin, TemporaryFile() as stdout:
            stdin.seek(0)
            proc = Popen(command, stdin=stdin, stdout=stdout, stderr=stdout, cwd=working_directory)
            try:
                proc.wait(timeout=timeout)
            except TimeoutExpired:
                proc.kill()
                raise TimedOut

            stdout.seek(0)
            output = stdout.read()

        split = output.decode('utf-8').split("\n")
        errors = []
        test_result = []
        preamble = True
        for line in split:
            if line.startswith("Error"):
                errors.append(line)
            if line.startswith("{"):
                preamble = False
            if preamble:
                continue
            test_result.append(line)

        test_result = "\n".join(test_result)

        if errors:
            print("Test output:")
            print(output.decode('utf-8'))
            raise TestRunException("\n".join(errors))
        try:
            return self._normalize_mocha(loads(test_result))
        except JSONDecodeError:
            print("Test output:")
            print(output.decode('utf-8'))
            raise TestRunException("Encountered error during test output analysis")

    def check_bytecodes(self, working_directory: str, original_bytecode: Dict[str, str]) -> bool:
        """ Returns whether none of the bytecodes differ from the original bytecodes

        :param working_directory: The truffle directory for which we should check the bytecodes
        :param original_bytecode: The original bytecodes {'contractName': '0x00'}
        :return: Whether the bytecodes match up
        """
        current_bytecodes = self.get_bytecodes(working_directory)
        for contractName, bytecode in current_bytecodes.items():
            if original_bytecode[contractName] != bytecode:
                return False
        return True

    def get_bytecodes(self, working_directory: str) -> Dict[str, str]:
        """ Returns the bytecodes in the compilation result of the current directory

        :param working_directory: The truffle directory for which we retreive the bytecodes
        :return: bytecodes in the shape {'contractName': '0x00'}
        """
        w_dir = Path(working_directory)
        self.run_compile_command(working_directory)
        if not (w_dir / "build").is_dir():
            logger.error("Compilation did not create build directory")

        contracts_dir = w_dir / "build" / "contracts"
        if not contracts_dir.is_dir():
            logger.error("No contracts directory found in build directory")

        current_bytecode = {}

        for contract in contracts_dir.iterdir():
            try:
                contract_compilation_result = json.loads(contract.read_text('utf-8'))
            except json.JSONDecodeError:
                logger.warning(f"Could not read compilation result for {contract.name}")
                continue

            current_bytecode[contract_compilation_result["contractName"]] = \
                self._strip_metadata(contract_compilation_result["bytecode"])
        return current_bytecode

    @staticmethod
    def _strip_metadata(bytecode: str) -> str:
        return swarm_hash_regex.sub("", bytecode)

    @staticmethod
    def _normalize_mocha(mocha_json: dict) -> Dict[str, TestResult]:
        tests = {}
        for failure in mocha_json["failures"]:
            tests[failure["fullTitle"]] = TestResult(failure["title"], failure["fullTitle"], failure["duration"], False)
        for success in mocha_json["passes"]:
            tests[success["fullTitle"]] = TestResult(success["title"], success["fullTitle"], success["duration"], True)
        return tests



from typing import List
from eth_vertigo.core.mutation import Mutation


class MutationFilter:
    """ A core filter provides an interface to selectively filter test cases"""

    def apply(self, mutations: List[Mutation]) -> List[Mutation]:
        """ Apply this filter to a list of mutations

        :param mutations: The mutations to filter
        :return: The resulting list
        """
        raise NotImplementedError



from eth_vertigo.mutator.mutator import Mutator
from eth_vertigo.mutator.truffle.solidity_file import SolidityFile
from eth_vertigo.mutator.solidity.solidity_mutator import SolidityMutator
from eth_vertigo.test_runner.truffle import TruffleRunnerFactory
from eth_vertigo.test_runner.exceptions import EquivalentMutant
from eth_vertigo.core import Mutation, MutationResult
from eth_vertigo.core.campaign import Campaign
from eth_vertigo.core.truffle.truffle_compiler import TruffleCompiler
from eth_vertigo.test_runner.exceptions import TestRunException, TimedOut

from typing import List, Callable
from time import time
from pathlib import Path
import logging
from queue import Queue


class TruffleCampaign(Campaign):
    """
    A TruffleCampaign class

    Implements specific campaign logic for the truffle framework
    """
    def __init__(
            self,
            project_directory: Path,
            truffle_compiler: TruffleCompiler,
            truffle_runner_factory: TruffleRunnerFactory,
            mutators: List[Mutator],
            networks: List[str] = (),
            filters=None
    ):
        super().__init__(filters=filters)
        self.project_directory = project_directory
        self.source_directory = project_directory / "build" / "contracts"
        self.truffle_compiler = truffle_compiler

        self.sources = list(self._get_sources())
        self.base_run_time = None
        self.networks = networks
        self.networks_queue = Queue(maxsize=len(networks))
        self.bytecodes = {}

        self.truffle_runner_factory = truffle_runner_factory
        self.mutators = mutators
        self.mutators.append(SolidityMutator())

        for network in networks:
            self.networks_queue.put(network)

    def _get_sources(self, dir=None):
        """ Implements basic mutator file discovery """
        if not (self.project_directory / "build").exists():
            self.truffle_compiler.run_compile_command(str(self.project_directory))

        dir = dir or self.source_directory
        for source_file in dir.iterdir():
            if source_file.name == "Migrations.json":
                continue
            if not source_file.name.endswith(".json"):
                continue
            yield SolidityFile(source_file)

    def valid(self):
        """ Checks whether the current project is valid """
        tr = self.truffle_runner_factory.create(str(self.project_directory))

        begin = time()

        network = None
        if self.networks:
            network = self.networks_queue.get()
        try:
            test_result = tr.run_tests(network=network)
        finally:
            if self.networks:
                self.networks_queue.put(network)

        self.base_run_time = time() - begin
        if test_result is None:
            return False

        return all([result.success for result in test_result.values()])

    def setup(self):
        for source in self.sources:
            for mutator in self.mutators:
                self.mutations += mutator.mutate(source, self.project_directory)
        for f in self.filters:
            self.mutations = f.apply(self.mutations)
        self.is_set_up = True

    def test_mutation(self, mutation: Mutation, done_callback: Callable):
        """ Run the test suite using a core and check for murders """
        tr = self.truffle_runner_factory.create(str(self.project_directory))
        mutation.result = MutationResult.LIVED
        network = None
        if self.networks:
            network = self.networks_queue.get()
        try:
            try:
                test_result = tr.run_tests(
                    mutation=mutation,
                    timeout=int(self.base_run_time) * 2,
                    network=network,
                    original_bytecode=self.bytecodes
                )
                if any(map(lambda t: not t.success, test_result.values())):
                    mutation.result = MutationResult.KILLED
            except EquivalentMutant:
                mutation.result = MutationResult.EQUIVALENT
        except TimedOut:
            mutation.result = MutationResult.TIMEDOUT
        except TestRunException as e:
            logging.warning(str(e))
            mutation.result = MutationResult.ERROR
        finally:
            if self.networks:
                self.networks_queue.put(network)
            done_callback()
            return

    def store_compilation_results(self):
        """ Stores compilation results for trivial compiler equivalence"""
        self.bytecodes = self.truffle_compiler.get_bytecodes(working_directory=str(self.project_directory))



from eth_vertigo.mutator.source_file import SourceFile
from pathlib import Path


class Mutator:
    """ Mutator

    A mutator implements the logic to generate mutants
    """
    def mutate(self, source_file: SourceFile, project_directory: Path):
        raise NotImplementedError



from eth_vertigo.mutator.solidity.solidity_file import SolidityFile
from eth_vertigo.mutator.source_mutator import SourceMutator
from eth_vertigo.core.mutation import Mutation
from pathlib import Path

conditional_boundary_mutator = {
    "==": " != ",
    "<": " <= ",
    "<=": " < ",
    ">": " >= "
}

conditional_negation_mutator = {
    "==": " != ",
    "!=": " == ",
    "<=": " > ",
    ">=": " < ",
    "<": " >= ",
    ">": " <= "
}

increments_mutator = {
    "+=": " -= ",
    "-=": " += "
}

math_mutants = {
    "+": " - ",
    "-": " + ",
    "*": " / ",
    "/": " * ",
    "&": " | ",
    "|": " & ",
    "^": " & ",
    "~": "",
    "<<": " >> ",
    ">>": " << "
}

mirror_mutants = {
    "-=": "=-",
    "+=": "=+"
}


def _mutate_binary_op(
    mutate_dict: dict, source_file: SolidityFile, project_directory: Path
):
    interesting_locs = source_file.get_binary_op_locations()

    for original_operator, src in interesting_locs:
        if original_operator not in mutate_dict.keys():
            continue
        yield Mutation(
            src, source_file, mutate_dict[original_operator], project_directory
        )


def _mutate_assignment(
    mutate_dict: dict, source_file: SolidityFile, project_directory: Path
):
    interesting_locs = source_file.get_assignments()

    for original_operator, src in interesting_locs:
        if original_operator not in mutate_dict.keys():
            continue
        yield Mutation(
            src, source_file, mutate_dict[original_operator], project_directory
        )


class SolidityMutator(SourceMutator):
    @staticmethod
    def mutate_boundary_conditionals(source_file: SolidityFile, project_directory: Path):
        return _mutate_binary_op(conditional_boundary_mutator, source_file, project_directory)

    @staticmethod
    def mutate_negated_conditionals(source_file: SolidityFile, project_directory: Path):
        return _mutate_binary_op(conditional_negation_mutator, source_file, project_directory)

    @staticmethod
    def mutate_math_ops(source_file: SolidityFile, project_directory: Path):
        return _mutate_binary_op(math_mutants, source_file, project_directory)

    @staticmethod
    def mutate_increments(source_file: SolidityFile, project_directory: Path):
        return _mutate_assignment(increments_mutator, source_file, project_directory)

    @staticmethod
    def mutate_mirror(source_file: SolidityFile, project_directory: Path):
        return _mutate_assignment(mirror_mutants, source_file, project_directory)

    @staticmethod
    def mutate_voids(source_file: SolidityFile, project_directory: Path):
        void_calls = list(source_file.get_void_calls())
        for _, src in void_calls:
            src[1] += 1
            yield Mutation(src, source_file, "", project_directory)

    @staticmethod
    def mutate_modifier(source_file: SolidityFile, project_directory: Path):
        modifier_invocations = list(source_file.get_modifier_invocations())
        for _, src in modifier_invocations:
            yield Mutation(src, source_file, "", project_directory)



from enum import Enum


class MutationStrategy(Enum):
    """
    Mutation strategy enum that signifies the different strategies that are available
    """
    BoundaryConditionals = 1
    NegatedConditionals = 2
    MathOperations = 3
    MutateIncrements = 4
    MutateVoidCalls = 5


class MutationConfiguration:
    """This class is used to convey information to the Mutator instances

    Selecting core profiles:
    This configuration object lets you configure a Mutator instance to only execute specific mutations strategies
    and leave others
    """

    def __init__(self, enabled_strategies=None):
        """ Initialize core configuration object

        :param enabled_strategies:
        """
        self.enabled_strategies = enabled_strategies or []

    @property
    def all_enabled(self):
        """ Returns whether all strategies are enabled at this moment"""
        return len(self.enabled_strategies) in (0, 5)

    def is_enabled(self, strategy: MutationStrategy):
        """ Returns whether the given strategy is enabled in this configuration"""
        return strategy in self.enabled_strategies



from typing import List
from eth_vertigo.core.mutation import Mutation
from eth_vertigo.core.filter import MutationFilter
from random import sample


class SampleFilter(MutationFilter):
    """ A sample filter which takes a random subset of the mutations along a defined ratio"""

    def __init__(self, ratio: float = 0.1):
        """ Creates a sample filter

        :param ratio: ratio of mutations to select
        """
        self.ratio = ratio
        super().__init__()

    def apply(self, mutations: List[Mutation]) -> List[Mutation]:
        """ Apply this filter to a list of mutations

        :param mutations: The mutations to filter
        :return: The resulting list
        """
        return sample(mutations, int(self.ratio * len(mutations)))



from eth_vertigo.mutator.universal_mutator.rule import Rule
from pathlib import Path
from typing import Generator
from re import compile, error
from loguru import logger


class RuleLoader:
    """ RuleLoader

    This class implements the logic to load universal mutator style mutation rules
    """

    def load_from_file(self, file: Path) -> Generator[Rule, None, None]:
        """Load mutation rules from file"""
        file_contents = file.read_text(encoding='utf-8')
        return self.load_from_txt(file_contents)

    def load_from_txt(self, rule_txt: str) -> Generator[Rule, None, None]:
        """Load mutation rules from text"""
        for line in rule_txt.split("\n"):
            # Deal with non mutation rule lines
            if "==>" not in line:
                if line.startswith("#"):
                    continue
                if line == "":
                    continue
                logger.warning(f"Error while parsing Universal Mutator rules:\n{line}")
                continue

            split_line = line.split("==>")
            if len(line) < 2:
                logger.warning(f"Error while parsing Universal Mutator rules:\n{line}")
                continue

            match = split_line[0]
            replace = split_line[1]

            # compile regex
            try:
                match = compile(match)
            except error:
                logger.warning(f"Error while parsing Universal Mutator rules.\nInvalid regular expression:\n{line}")

            # drop return char if present
            if replace and replace[-1] == '\n':
                replace = replace[:-1]
                replace.strip(' ')
            yield Rule(match, replace)



from typing import List
from eth_vertigo.core.mutation import Mutation
from eth_vertigo.core.filter import MutationFilter
from random import sample


class ExcludeFilter(MutationFilter):
    """ A exclusion filter to ignore files from specific directories """

    def __init__(self, prefixes: List[str]):
        """ Creates an exclusion filter """
        self.prefixes = prefixes
        super().__init__()

    def apply(self, mutations: List[Mutation]) -> List[Mutation]:
        """ Apply this filter to a list of mutations

        :param mutations: The mutations to filter
        :return: The resulting list
        """

        def should_not_filter(mutation):
            for prefix in self.prefixes:
                if str(mutation.relative_path).startswith(prefix) or str(mutation.relative_path).startswith("/" + prefix):
                    return False
            return True

        return list(filter(should_not_filter, mutations))



from pathlib import Path
from typing import Generator
from eth_vertigo.core import Mutation
from eth_vertigo.mutator.source_file import SourceFile
from re import finditer


class Rule:
    """ Rule

    This class represents a single universal mutator style mutation rule.

    Additionally the class provides logic to apply the rule to smart contracts.
    """

    def __init__(self, match, replace: str):
        """ Instantiate a new Rule

        :param match: The regex expression that specifies which parts of a program to replace
        :param replace: The string with which to replace the matches
        """
        self.match = match
        self.replace = replace

    def generate_mutants(self, source: SourceFile, project_directory: Path) -> Generator[Mutation, None, None]:
        if self.replace in ("DO_NOT_MUTATE", ):
            return

        file = source.file
        file_content = file.read_text(encoding="utf-8")

        for occurrence in finditer(self.match, file_content):
            start = occurrence.start()
            end = occurrence.end() - 1
            size = end - start
            yield Mutation((start, size, 0), source, self.replace, project_directory)



from pathlib import Path
from eth_vertigo.mutator.source_file import SourceFile
from typing import Tuple
from enum import Enum
from jinja2 import PackageLoader, Environment

environment = Environment(
        loader=PackageLoader("eth_vertigo.core"), trim_blocks=True
    )

class MutationResult(Enum):
    KILLED = 1
    LIVED = 2
    TIMEDOUT = 3
    ERROR = 4
    EQUIVALENT = 5


_mutationresult_string = {
    MutationResult.KILLED: "Killed",
    MutationResult.LIVED: "Lived",
    MutationResult.TIMEDOUT: "Timeout",
    MutationResult.ERROR: "Error",
    MutationResult.EQUIVALENT: "Equivalent",
    None: "None"
}


class Mutation:
    """
    A core class contains specific information on a single core
    """

    def __init__(self, location: Tuple[int, int, int], source: SourceFile, value: str, project_directory: Path):
        """
        Initializes a core
        :param location: Location of the core, in the src format (offset, length, file_index)
        :param source: Source file for which the core is to be applied
        :param value: New value that the location should take on
        :param project_directory: Project directory Path of the project directory that eth_vertigo is working in
        """
        self.location = location
        self.source = source
        self.project_directory = project_directory
        self.value = value

        # The following parameters are used to track how and when this core was killed
        self.result = None
        self.crime_scenes = []

    @property
    def relative_path(self):
        """ Gets the relative path of the mutator wrt the project directory """
        r_path = self.source.file.relative_to(self.project_directory)
        return str(r_path)

    @staticmethod
    def _get_mutated_line(offset, text):
        cursor = 0
        for i, line in enumerate(text.splitlines(keepends=True)):
            if len(line) + cursor > offset:
                return i, line.replace("\t", "")
            cursor += len(line)

    def __repr__(self) -> str:
        """ Prints information that can be used to triage a core """
        template = environment.get_template("mutation_template.jinja2")
        source_content = self.source.file.read_text('utf-8')
        line_nr, og_line = self._get_mutated_line(self.location[0], source_content)

        mutated = source_content[:self.location[0]] \
                  + self.value + \
                  source_content[self.location[0] + self.location[1]:]

        _, mut_line = self._get_mutated_line(self.location[0], mutated)

        return template.render(
            file_name=str(self.source.file),
            line_number=line_nr,
            original_line=og_line,
            mutated_line=mut_line,
            result=_mutationresult_string[self.result]
        )



from pathlib import Path
from typing import Tuple


class SourceFile:
    """ SourceFile objects represent contract files

    Specifically it stores information on the file it represents and it exposes functions that
    allow other components to find specific interesting locations in the mutator file.
    This is mostly used by the different Mutator implementations
    """

    def __init__(self, file: Path):
        self.file = file

    def get_binary_op_locations(self) -> Tuple[str, Tuple[int, int, int]]:
        """Gets locations for all the binary operations in the sourcefile
        :returns tuple with (original_operation, src)
        """
        raise NotImplementedError

    def get_if_statement_binary_ops(self) -> Tuple[str, Tuple[int, int, int]]:
        """Gets locations for all the binary operations in the sourcefile

        Specifically those supplied as an argument to an if statement
        :returns tuple with (original_operation, src)
        """
        raise NotImplementedError

    def get_assignments(self) -> Tuple[str, Tuple[int, int, int]]:
        """Gets locations for all the assignments in the sourcefile
        :returns tuple with (original_operation, src)
        """
        raise NotImplementedError

    def get_void_calls(self) -> Tuple[str, Tuple[int, int, int]]:
        """Gets locations for all the void calls in the sourcefile
        :returns tuple with (original_operation, src)
        """
        raise NotImplementedError



from eth_vertigo.test_runner.test_result import TestResult
from typing import Dict, Union


class TruffleTester:
    """Truffle tester interface exposes testing functionality from truffle"""
    def run_test_command(self, working_directory: str, file_name: str = None, network_name: str = None, timeout=None) \
            -> Union[None, Dict[str, TestResult]]:
        """ Runs truffle's test command

        :param working_directory: The directory which will be tested
        :param file_name: Filename of the test that should be executed
        :param network_name: Name of the network to execute the tests on
        :param timeout: The maximum duration that the tests are allowed to take
        :return: The test results
        """
        raise NotImplementedError

    def check_bytecodes(self, working_directory:str, original_bytecode: Dict[str, str]) -> bool:
        """ Returns whether any of the bytecodes differ from the original bytecodes

        :param working_directory: The truffle directory for which we should check the bytecodes
        :param original_bytecode: The original bytecodes {'contractName': '0x00'}
        :return: Whether the bytecodes match up
        """
        raise NotImplementedError
